from itertools import islice

MAX_DISPLAY = 100


def get_keys_and_children(inspector):
    # positron frontend only displays the first 100 items, then the count of the rest.
    keys, values = [], []
    keys_iterator = iter(inspector.get_children())
    for key in islice(keys_iterator, MAX_DISPLAY):
        keys.append(str(key))
        try:
            val = inspector.get_child(key)
        except Exception:
            # TODO: ideally this should be a sentiel value indicating it's
            # not a python string, but the actual display value we want. For now,
            # at least indicate some error.
            val = "<error>"
        values.append(val)

    if len(keys) == MAX_DISPLAY:
        # check if there are more children
        n_children = inspector.get_length()
        if n_children == 0:
            # no len() method, finish iteratoring over keys_iterator to get the true size
            for n_children, _ in enumerate(keys_iterator, MAX_DISPLAY + 1):
                pass
        n_remaining = n_children - MAX_DISPLAY
        if n_remaining > 0:
            keys.append("[[...]]")
            values.append(ChildrenOverflow(n_remaining))

    return keys, values


def get_child(inspector, index):
    key = next(islice(inspector.get_children(), index - 1, None))
    return inspector.get_child(key)


class ChildrenOverflow:
    def __init__(self, n_remaining):
        self.n_remaining = n_remaining



import sys
import types
import inspect


def isstring(s):
    # if we use Python 3
    if sys.version_info[0] >= 3:
        return isinstance(s, str)
    # we use Python 2
    return isinstance(s, basestring)


def normalize_func(func):
    # return None for builtins
    if inspect.isbuiltin(func):
        return None
    return func


def get_doc(func):
    doc = inspect.getdoc(func)
    if doc is None:
        func = normalize_func(func)
        if func is None:
            return None
        else:
            doc = inspect.getdoc(func)
    return doc


def get_property_doc(target, prop):
    for name, obj in inspect.getmembers(type(target), inspect.isdatadescriptor):
        if isinstance(obj, property) and name == prop:
            return inspect.getdoc(obj.fget)
    return None


def get_argspec(func):
    try:
        if sys.version_info[0] >= 3:
            return inspect.getfullargspec(func)
        else:
            return inspect.getargspec(func)
    except TypeError:
        return None


def get_arguments(func):
    func = normalize_func(func)
    if func is None:
        return None
    argspec = get_argspec(func)
    if argspec is None:
        return None
    args = argspec.args
    if "self" in args:
        args.remove("self")
    return args


def get_r_representation(default):
    if callable(default) and hasattr(default, "__name__"):
        arg_value = default.__name__
    else:
        if default is None:
            arg_value = "NULL"
        elif type(default) == type(True):
            if default == True:
                arg_value = "TRUE"
            else:
                arg_value = "FALSE"
        elif isstring(default):
            arg_value = '"%s"' % default
        elif isinstance(default, int):
            arg_value = "%rL" % default
        elif isinstance(default, float):
            arg_value = "%r" % default
        elif isinstance(default, list):
            arg_value = "c("
            for i, item in enumerate(default):
                if i is (len(default) - 1):
                    arg_value += "%s)" % get_r_representation(item)
                else:
                    arg_value += "%s, " % get_r_representation(item)
        elif isinstance(default, (tuple, set)):
            arg_value = "list("
            for i, item in enumerate(default):
                if i is (len(default) - 1):
                    arg_value += "%s)" % get_r_representation(item)
                else:
                    arg_value += "%s, " % get_r_representation(item)
        elif isinstance(default, dict):
            dict_entries = ", ".join(
                [
                    ("%s = %s" % (k, get_r_representation(v)))
                    for k, v in default.items()
                ]
            )
            arg_value = "list(" + dict_entries + ")"
        else:
            arg_value = "%r" % default

    # if the value starts with "tf." then convert to $ usage
    if arg_value.startswith("tf."):
        arg_value = arg_value.replace(".", "$")

    return arg_value


def generate_signature_for_function(func):
    """Given a function, returns a string representing its args."""

    func = normalize_func(func)
    if func is None:
        return None

    args_list = []

    argspec = get_argspec(func)
    if argspec is None:
        return None

    first_arg_with_default = len(argspec.args or []) - len(
        argspec.defaults or []
    )
    for arg in argspec.args[:first_arg_with_default]:
        if arg == "self":
            # Python documentation typically skips `self` when printing method
            # signatures.
            continue
        args_list.append(arg)

    if (
        argspec.varargs == "args"
        and hasattr(argspec, "keywords")
        and argspec.keywords == "kwds"
    ):
        original_func = func.__closure__[0].cell_contents
        return generate_signature_for_function(original_func)

    if argspec.defaults:
        for arg, default in zip(
            argspec.args[first_arg_with_default:], argspec.defaults
        ):
            arg_value = get_r_representation(default)
            args_list.append("%s = %s" % (arg, arg_value))
    if argspec.varargs:
        args_list.append("...")
    if hasattr(argspec, "keywords") and argspec.keywords:
        args_list.append("...")
    return "(" + ", ".join(args_list) + ")"



# NOTE: indirection required here for Python 2.7 support
# TypeError: unicode argument expected, got 'str'
import sys

if sys.version_info < (3, 0):
    from io import BytesIO as StringIO
else:
    from io import StringIO


def _setStream(handler, stream):
    setStream = getattr(handler, "setStream", None)
    if setStream is not None:
        return setStream(stream)
    old_stream = handler.stream
    handler.stream = stream
    return old_stream


def _override_logger_streams(
    capture_stdout,
    old_stdout,
    new_stdout,
    capture_stderr,
    old_stderr,
    new_stderr,
):
    import logging

    # capture root handlers
    root = getattr(logging, "root", None)
    if root is not None:
        handlers = getattr(root, "handlers", [])
        for handler in handlers:

            stream = getattr(handler, "stream", None)
            if stream is None:
                continue
            if capture_stdout and stream is old_stdout:
                _setStream(handler, new_stdout)
            elif capture_stderr and stream is old_stderr:
                _setStream(handler, new_stderr)
    # capture loggers registered with the default manager
    loggers = getattr(logging.Logger.manager, "loggerDict", {})
    for logger in loggers.values():
        handlers = getattr(logger, "handlers", [])
        for handler in handlers:
            stream = getattr(handler, "stream", None)
            if stream is None:
                continue
            if capture_stdout and stream is old_stdout:
                _setStream(handler, new_stdout)
            elif capture_stderr and stream is old_stderr:
                _setStream(handler, new_stderr)


class OutputCaptureContext:
    def __init__(self, capture_stdout, capture_stderr):
        self._capture_stdout = bool(capture_stdout)
        self._capture_stderr = bool(capture_stderr)
        self._capturing_stream = StringIO()

    def __enter__(self):
        if self._capture_stdout:
            self._prev_stdout = sys.stdout
            sys.stdout = self._capturing_stream

        if self._capture_stderr:
            self._prev_stderr = sys.stderr
            sys.stderr = self._capturing_stream

        _override_logger_streams(
            capture_stdout=self._capture_stdout,
            new_stdout=sys.stdout if self._capture_stdout else None,
            old_stdout=self._prev_stdout if self._capture_stdout else None,
            capture_stderr=self._capture_stderr,
            new_stderr=sys.stderr if self._capture_stderr else None,
            old_stderr=self._prev_stderr if self._capture_stderr else None,
        )

    def __exit__(self, *args):
        if self._capture_stdout:
            sys.stdout = self._prev_stdout
        if self._capture_stderr:
            sys.stderr = self._prev_stderr

        _override_logger_streams(
            capture_stdout=self._capture_stdout,
            new_stdout=sys.stdout,
            old_stdout=self._prev_stdout if self._capture_stdout else None,
            capture_stderr=self._capture_stderr,
            new_stderr=sys.stderr,
            old_stderr=self._prev_stderr if self._capture_stderr else None,
        )

    def collect_output(self, clear=True):
        output = self._capturing_stream.getvalue()
        if clear:
            self._capturing_stream.truncate(0)
            self._capturing_stream.seek(0)
        return output


class OutputRemap(object):

    def __init__(self, target, handler, tty=True):
        self.target = target
        self.handler = handler
        self.tty = tty

    def write(self, message):
        return self.handler(message)

    def isatty(self):
        return self.tty

    def __getattr__(self, attr):
        if self.target:
            return getattr(self.target, attr)
        else:
            return 0

    def close(self):
        return None

    def flush(self):
        return None


def _remap_output_streams(r_stdout, r_stderr, tty):
    sys.stdout = OutputRemap(sys.stdout, r_stdout, tty)
    sys.stderr = OutputRemap(sys.stderr, r_stderr, tty)


class RemapOutputStreams:
    def __init__(self, r_stdout, r_stderr, tty):
        self.r_stdout = r_stdout
        self.r_stderr = r_stderr
        self.tty = tty
        self._stdout = sys.stdout
        self._stderr = sys.stderr

    def __enter__(self):
        # It's possible that __enter__ does not execute before __exit__ in some
        # special cases. We also store _stdout and _stderr when creating the context.
        self._stdout = sys.stdout
        self._stderr = sys.stderr

        _remap_output_streams(self.r_stdout, self.r_stderr, self.tty)

    def __exit__(self, *args):
        sys.stdout = self._stdout
        sys.stderr = self._stderr



import sys
import os


def run_file(path):
    with open(path, "r") as file:
        file_content = file.read()

    # ipykernel patches the loader, so that
    # `import __main__` does not produce the "real" __main__, but rather,
    # the facade that is the user facing __main__
    # to get the "real" main, do:
    # d = sys.modules["__main__"].__dict__
    from __main__ import __dict__ as d

    exec(file_content, d, d)


class RunMainScriptContext:
    def __init__(self, path, argv=None):
        self.path = path
        self.argv = tuple(argv) if argv is not None else None

    def __enter__(self):
        sys.path.insert(0, os.path.dirname(self.path))

        if self.argv is not None:
            self._orig_sys_argv = sys.argv
            sys.argv = [self.path] + list(self.argv)

    def __exit__(self, *_):
        # try restore sys.path
        try:
            sys.path.remove(os.path.dirname(self.path))
        except ValueError:
            pass
        # try restore sys.argv if we patched it
        if self.argv is not None:
            # restore sys.argv if it's unmodified from what we set it to.
            # otherwise, leave it as-is.
            patched_argv = [self.path] + list(self.argv)
            if sys.argv == patched_argv:
                sys.argv = self._orig_sys_argv


def _launch_lsp_server_on_thread(path, args):
    # used by Positron reticulate launcher...
    # TODO: update Positron to replace usage of this with `run_file_on_thread()`

    return run_file_on_thread(path, args)


def set_blank_io_streams():
    import sys
    import os
    sys.stdout = open(os.devnull, "w")
    sys.stderr = open(os.devnull, "w")


def run_file_on_thread(path, argv=None, init_globals=None, run_name="__main__"):
    # for now, leave sys.argv and sys.path permanently modified.
    # Later, revisit if it's desirable/safe to restore after the initial
    # lsp event loop startup.
    import _thread
    from runpy import run_path

    RunMainScriptContext(path, argv).__enter__()
    _thread.start_new_thread(
        run_path,
        (path,),
        {
            "run_name": run_name,
            "init_globals": init_globals,
        },
    )



import threading
import collections

import sys

is_py2 = sys.version[0] == "2"
if is_py2:
    import Queue as queue
else:
    import queue as queue


def isScalar(x):
    return not isinstance(x, (list, tuple))


def isList(x):
    return isinstance(x, (list))


def asString(x):
    return str(x)


def makeDict():
    return {"a": 1.0, "c": 3.0, "b": 2.0}


def makeTuple():
    return (1.0, 2.0, 3.0)


def makeTupleWithOrderedDict():
    return (1.0, collections.OrderedDict({"b": 777, "a": 22}))


def makeIterator(x):
    return iter(x)


def makeGenerator(n):
    i = 0
    while i < n:
        yield i
        i += 1


def iterateOnThread(iter):
    results = []

    def iteration_worker():
        for i in iter:
            results.append(i)

    thread = threading.Thread(target=iteration_worker)
    thread.start()
    while thread.is_alive():
        thread.join(0.1)
    return results


def invokeOnThread(f, *args, **kwargs):
    result = []

    def invoke_worker():
        result.append(f(*args, **kwargs))

    thread = threading.Thread(target=invoke_worker)
    thread.start()
    while thread.is_alive():
        thread.join(0.1)
    return result[0]


def reflect(x):
    return x


def callFunc(f, *args, **kwargs):
    return f(*args, **kwargs)


def testThrowError():
    throwError()


def throwError():
    raise ValueError("A very specific bad thing happened")


class PythonClass(object):

    FOO = 1
    BAR = 2

    @classmethod
    def class_method(cls):
        return cls.FOO


class PythonCallable(object):

    FOO = 1
    BAR = 2

    """ Call a callable
    Args:
      arg1: First argument.
  """

    def __call__(self, arg1):
        return arg1


def create_callable():
    return PythonCallable()


dict_with_callable = dict(callable=create_callable())



from rpycall import call_r_function


def make_python_function(f, name=None):
    def python_function(*args, **kwargs):
        return call_r_function(f, *args, **kwargs)

    if name is not None:
        python_function.__name__ = name

    return python_function



import sys
import threading

if sys.version_info.major < 3:
    import __builtin__ as builtins
else:
    import builtins


# The R callback to be run. Initialized in the 'initialize()' method.
_callback = None

# A list of Python packages which have been imported. The aforementioned
# callback will be run on the main thread after a module has been imported
# on the main thread.
_imported_packages = []

# A simple counter, tracking the recursion depth. This is used as we only
# attempt to run the R callback at the top level; that is, we don't want
# to run it while modules are being loaded recursively.
_recursion_depth = 0

# The builtin implementation of '__import__'; saved so that we can re-use it
# after initialization.
__import__ = builtins.__import__

# The implementation of '_find_and_load' captured from 'importlib._bootstrap'.
# Since we're trying to poke at Python internals, we try to wrap this code
# in try-catch and only use this if it appears safe to do so.
_find_and_load = None
try:
    import importlib._bootstrap

    _find_and_load = importlib._bootstrap._find_and_load
except:
    pass


# Run hooks on imported packages, if safe to do so.
def _maybe_run_hooks():

    # Don't run hooks while loading packages recursively.
    global _recursion_depth
    if _recursion_depth != 0:
        return False

    # Check whether we're on the main thread. Note that separate threads can
    # attempt to load Python modules, but the R callback we register can only
    # be safely run on the main thread.
    is_main_thread = isinstance(
        threading.current_thread(), threading._MainThread
    )
    if not is_main_thread:
        return False

    # Pre-flight checks passed; run the callbacks.
    global _imported_packages
    global _callback
    for package in _imported_packages:
        _callback(package)

    # Clear the import list.
    del _imported_packages[:]


# Resolve a module name on import. See Python code here for motivation.
# https://github.com/python/cpython/blob/c5140945c723ae6c4b7ee81ff720ac8ea4b52cfd/Lib/importlib/_bootstrap.py#L1246-L1270
def _resolve_module_name(name, globals=None, level=0):

    if level == 0:
        return name

    package = globals.get("__package__")
    if package is not None:
        return package

    spec = globals.get("__spec__")
    if spec is not None:
        return spec.parent

    return name


# Helper function for running an import hook with our extra scaffolding.
def _run_hook(name, hook):

    # Check whether this module has already been imported.
    already_imported = name in sys.modules

    # Bump the recursion depth.
    global _recursion_depth
    _recursion_depth += 1

    # Run the hook.
    try:
        module = hook()
    except:
        raise
    finally:
        _recursion_depth -= 1

    # Add this package to the import list, if this is the first
    # time importing that package.
    global _imported_packages
    if not already_imported:
        _imported_packages.append(name)

    # try and run hooks if possible
    _maybe_run_hooks()

    # return loaded module
    return module


# The hook installed to replace 'importlib._bootstrap._find_and_load'.
def _find_and_load_hook(name, import_):

    def _hook():
        global _find_and_load
        return _find_and_load(name, import_)

    return _run_hook(name, _hook)


# Initialize the '_find_and_load' replacement hook.
def _initialize_importlib():

    import importlib._bootstrap

    importlib._bootstrap._find_and_load = _find_and_load_hook


# The hook installed to replace '__import__'.
def _import_hook(name, globals=None, locals=None, fromlist=(), level=0):

    # resolve module name
    resolved_module_name = _resolve_module_name(name, globals, level)

    def _hook():
        global __import__
        return __import__(
            name, globals=globals, locals=locals, fromlist=fromlist, level=level
        )

    return _run_hook(_hook)


# Initialize the '__import__' hook.
def _initialize_default():
    builtins.__import__ = _import_hook


# The main entrypoint for this module.
def initialize(callback):

    # Save the callback.
    global _callback
    _callback = callback

    # Check whether we can initialie with importlib.
    global _find_and_load
    if _find_and_load is not None:
        return _initialize_importlib()

    # Otherwise, fall back to default implementation.
    return _initialize_default()



import sys
import queue as queue
from rpycall import call_r_function, schedule_python_function_on_main_thread


def safe_call_r_function(f, args, kwargs):
    try:
        return call_r_function(f, *args, **kwargs), None
    except Exception as e:  # TODO: should we catch BaseException too? KeyboardInterrupt?
        return None, e


def safe_call_r_function_on_main_thread(f, *args, **kwargs):
    result = queue.Queue()
    schedule_python_function_on_main_thread(
        lambda: result.put(safe_call_r_function(f, args, kwargs)),
        None,
    )
    return result.get()
